:- multifile portray/1.

:- (clpfd:full_answer -> true ; assert(clpfd:full_answer)).
:- use_module(library(process)).

:- dynamic(coloring/1).

coloring(on).
setcolor(X):- retract_all(coloring(_)),assert(color(X)).

% Prints strings as "ciao" instead of [99,105,97,111]
portray(L):-
    is_code_list(L),
    length(L,Len),
    %prolog_flag(debugger_print_options,O,O),
    %memberchk(print_depth(Depth),O)
    Depth=10,   % Semplifichiamo, mettiamo la Depth di default
    (Depth<Len -> format('"~10s..."',[L])
        ; format('~s',[L])
    ).

portray(ic(Body,Head)):-
    write('ic:'),
    nl,
    print_body(Body),
    nl,
    write('--->\t'),
    print_disjunction(Head), nl.

portray(psic(Body,Head)):-
    write('psic:'),
    nl,
    print_body(Body),
    nl,
    write('--->\t'),
    print_disjunction(Head), nl.

portray(e([F,N],A,T)):-
    functor(A,F,N),
    print(e(A,T)).
portray(h([F,N],A,T)):-
    functor(A,F,N),
    color(blue),
    print(h(A,T)),
    color(white).
portray(en([F,N],A,T)):-
    functor(A,F,N),
    color(green),
    print(en(A,T)),
    color(white).


%    portray(forallf(X)):-
%        put_code(31), % Code of the forall symbol: a black triangle pointing down
%        print(X).


% CLP constraints


    portray(clpfd:'t=u IND'(A,B)):-
        print(A), write(#=), print(B).

    portray(clpfd:'x\\=y IND'(A,B)):-
        print(A), write(#\=), print(B).


    portray(clpfd:'t=<u+c'(A,B,C)):-
        (C=0    -> print(A#=<B);
         C= -1   -> print(A#<B);
         C>0    -> print(A#=<B+C);
                   print(A#=<B), print(C) % Negative number: sign written by print(C)
         ).


    portray(clpfd:'t>=u+c'(A,B,C)):-
        (C=0    -> print(A#>=B);
         C=1    -> print(A#>B);
         C>0    -> print(A#>=B+C);
                   print(A#>=B), print(C) % Negative number: sign written by print(C)
         ).

%%% Auxiliary predicates
is_code_list([]).
is_code_list([H|T]):-
    integer(H),
    is_code_list(T).

print_disjunction([H]):- !,
    write(H).
print_disjunction([H1,H2|T]):-
    write(H1), nl, write('\\/\t'),
    print_disjunction([H2|T]).

print_body([]).
print_body([H|T]):- !,
    print_body(H),
    print_body(T).
print_body(A):- print(A).

print_quantifiers(A):-
    var(A), get_quant(A,Q),!, Quant =.. [Q,A], print(Quant).
print_quantifiers(A):- var(A),!,print(unquant(A)).
print_quantifiers(A):- ground(A),!,print(ground(A)).
print_quantifiers([A|T]):-
    print_quantifiers(A),
    print_quantifiers(T).
print_quantifiers(A):-
    A =.. [_|ArgList],
    print(arglist(ArgList)),
    print_quantifiers(ArgList).

color(_):-coloring(off),!.

color(grey):- absolute_file_name(path('bash.exe'),F,[access(exist)]),process_create(F,['-c',[echo,' -e ','"\e[1;30m"']]).
color(cyan):- absolute_file_name(path('bash.exe'),F,[access(exist)]),process_create(F,['-c',[echo,' -e ','"\e[1;36m"']]).
color(white):- absolute_file_name(path('bash.exe'),F,[access(exist)]),process_create(F,['-c',[echo,' -e ','"\e[1;37m"']]).
color(blue):- absolute_file_name(path('bash.exe'),F,[access(exist)]),process_create(F,['-c',[echo,' -e ','"\e[1;34m"']]).
color(yellow):- absolute_file_name(path('bash.exe'),F,[access(exist)]),process_create(F,['-c',[echo,' -e ','"\e[1;33m"']]).
color(green):- absolute_file_name(path('bash.exe'),F,[access(exist)]),process_create(F,['-c',[echo,' -e ','"\e[1;32m"']]).
color(red):- absolute_file_name(path('bash.exe'),F,[access(exist)]),process_create(F,['-c',[echo,' -e ','"\e[1;31m"']]).
